#include "archive.h"
#include "extapp_api.h"
#include "../global_preferences.h"

#include <string.h>
#include <stdlib.h>

namespace External {
  namespace Archive {

#ifdef DEVICE

    static_assert(sizeof(TarHeader) == 512);

    bool isSane(const TarHeader* tar) {
      unsigned char ch=tar->name[0];
      return !memcmp(tar->magic, "ustar  ", 8) &&  ch>=32 && ch<128;
    }

    bool isExamModeAndFileNotExecutable(const TarHeader* tar) {
      return GlobalPreferences::sharedGlobalPreferences()->isInExamMode() && (tar->mode[4] & 0x01) == 0;
    }

    const int last_certified_index=8; // files 0 to 8 are certified

    bool ck_exammode(const TarHeader* tar,int index) {
      return GlobalPreferences::sharedGlobalPreferences()->isInExamMode() && index>last_certified_index;
    }

    int tar_size( const TarHeader* tar){
      if (!isSane(tar)) 
	return -1;
      size_t size = 0;
      for (int i=0; i<11; i++){
	char ch=tar->size[i];
	if (ch==' ')
	  return size;
	if (ch<'0' || ch>'7')
	  return -1;
	size = size*8 + (ch-'0');
      }
      return size;
    }

    void tar_get_entry(const TarHeader* tar,size_t size,File & entry){
      entry.name = tar->name;
      entry.data = reinterpret_cast<const uint8_t*>(tar) + sizeof(TarHeader);
      entry.dataLength = size;
      entry.isExecutable = (tar->mode[4] & 0x01)==1;
      entry.readable = (tar->mode[4] & 0x04)==1;
    }

    const TarHeader* tar_nxt(const TarHeader* tar,size_t size){
      unsigned stride = (sizeof(TarHeader) + size + 511);
      stride = (stride >> 9) << 9;
      tar = reinterpret_cast<const TarHeader*>(reinterpret_cast<const char*>(tar) + stride);
      return tar;
    }    

    bool fileAtIndex(size_t index, File &entry) {
      if (index==(size_t) -1)
	return false;  
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000);
      // TAR file is a set of records aligned to 512 bytes boundary followed by data.
      for (size_t idx=0;(size_t) tar<0x90800000;++idx) {
	// Calculate the size
	int size=tar_size(tar);
	if (size<0)
	  return false;
	// Check if we found our file.
	if (index==idx) {
	  // If yes, check for sanity and for exam mode stuff
	  if (!isSane(tar) || (tar->mode[4] & 0x04)==0 ||
	      ck_exammode(tar,index)
	      // isExamModeAndFileNotExecutable(tar)
	      )
	    return false;
	  // File entry found, copy data out.
	  tar_get_entry(tar,size,entry);
	  return true;
	}
	// move to the next TAR header.
	tar=tar_nxt(tar,size);
      }
  
      // Achievement unlock: How did we get there ?
      return false;
    }

    extern "C" void (* const apiPointers[])(void);
    typedef uint32_t (*entrypoint)(const uint32_t, const void *, void *, const uint32_t);

    uint32_t executeFile(const char *name, void * heap, const uint32_t heapSize) {
      File entry;
      int i=indexFromName(name);
      if(fileAtIndex(i, entry)) {
	if(!entry.isExecutable || i>last_certified_index) {
	  return 0;
	}
	uint32_t ep = *reinterpret_cast<const uint32_t*>(entry.data);
	if(ep >= 0x90200000 && ep < 0x90800000) {
	  return ((entrypoint)ep)(API_VERSION, apiPointers, heap, heapSize);
	}
      }
      return -1;
    }

#define EFFICIENT 1

    int indexFromName(const char *name) {
#ifdef EFFICIENT
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000);
      for (int idx=0;(size_t) tar<0x90800000U;++idx){
	if (!isSane(tar) ||
	    ck_exammode(tar,idx)
	    // isExamModeAndFileNotExecutable(tar)
	    ) 
	  return -1;
	// Calculate the size
	int size = tar_size(tar);
	if (size<0)
	  return -1;
	if ( (tar->mode[4] & 0x04) && strcmp(name,tar->name)==0 )
	  return idx; // found
	// go to next header
	tar=tar_nxt(tar,size);
      }
      return -1;
#else // old inefficient code in O(n^2)
      File entry;

      for (int i = 0; fileAtIndex(i, entry); i++) {
	if (entry.readable && strcmp(name, entry.name) == 0) {
	  return i;
	}
      }

      return -1;
#endif
    }

    size_t numberOfFiles() {
#ifdef EFFICIENT
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000);
      int idx;
      for (idx=0;(size_t) tar<0x90800000U;++idx){
	if (!isSane(tar) ||
	    ck_exammode(tar,idx)
	    // isExamModeAndFileNotExecutable(tar)
	    ) 
	  return idx;
	// Calculate the size
	int size=tar_size(tar);
	if (size<0)
	  return idx;
	// go to next header
	tar=tar_nxt(tar,size);
      }
      return idx;
#else // old inefficient code in O(n^2)
      File dummy;
      size_t count;

      for (count = 0; fileAtIndex(count, dummy); count++);

      return count;
#endif
    }

    bool executableAtIndex(size_t index, File &entry) {
#ifdef EFFICIENT
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000);
      size_t idx=0;
      for (int count=0;(size_t) tar<0x90800000U;++count){
	if (!isSane(tar) ||
	    ck_exammode(tar,count)
	    // isExamModeAndFileNotExecutable(tar)
	    ) 
	  return false;
	// Calculate the size
	int size=tar_size(tar);
	if (size<0)
	  return false;
	if (tar->mode[4] & 0x01){
	  if (idx==index){
	    tar_get_entry(tar,size,entry);
	    return true;
	  }
	  idx++;
	}
	// go to next header
	tar=tar_nxt(tar,size);
      }
      return false;
#else // old inefficient code in O(n^2)
      File dummy;
      size_t count;
      size_t final_count = 0;
      // inefficient but it's not important here since exe will be at tar begin 
      for (count = 0; fileAtIndex(count, dummy); count++) {
	if (dummy.isExecutable) {
	  if (final_count == index) {
	    entry.name = dummy.name;
	    entry.data = dummy.data;
	    entry.dataLength = dummy.dataLength;
	    entry.isExecutable = dummy.isExecutable;
	    entry.readable = dummy.readable;
	    return true;
	  }
	  final_count++;
	}
      }
  
      return false;
#endif
    }

    size_t numberOfExecutables() {
#ifdef EFFICIENT
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000U);
      int idx=0;
      for (int count=0;(size_t) tar<0x90800000U;++count){
	if (!isSane(tar) ||
	    ck_exammode(tar,count)
	    // isExamModeAndFileNotExecutable(tar)
	    ) 
	  return idx;
	// Calculate the size
	int size=tar_size(tar);
	if (size<0)
	  return idx;
	if (tar->mode[4] & 0x01)
	  idx++;
	// go to next header
	tar=tar_nxt(tar,size);
      }
      return idx;
#else // old inefficient code in O(n^2)
      File dummy;
      size_t count;
      size_t final_count = 0;
      // inefficient code but not that important since all exe are at tar begin
      for (count = 0; fileAtIndex(count, dummy); count++){
	if (dummy.isExecutable)
	  final_count++;
	// else break; // all exe must be at tar begin
      }

      return final_count;
#endif
    }



#else // DEVICE

    bool fileAtIndex(size_t index, File &entry) {
      if (index != 0)
	return false;
  
      entry.name = "Built-in";
      entry.data = NULL;
      entry.dataLength = 0;
      entry.isExecutable = true;
      entry.readable = true;
      return true;
    }

    bool executableAtIndex(size_t index, File &entry) {
      return fileAtIndex(index, entry);
    }

    size_t numberOfExecutables() {
      return 1;
    }

    extern "C" void extapp_main(void);

    uint32_t executeFile(const char *name, void * heap, const uint32_t heapSize) {
      extapp_main();
      return 0;
    }

    int indexFromName(const char *name) {
      if (strcmp(name, "Built-in") == 0)
	return 0;
      else
	return -1;
    }

    size_t numberOfFiles() {
      return 1;
    }

#endif // DEVICE

  }
}
